#include "combine.h"
#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>

/*
 * Reduce-Scatter
 */

// Ring Reduce-Scatter

int ring_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size, int * recvcounts) {
  int tag = 0;
  int i,j;
  double *tmpbuf;
  double *tmpbuf2;
  MPI_Request request;
  MPI_Status status;
  int * startelem;
  int curpos;

  startelem = (int*)malloc(size*sizeof(double));

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  startelem[0] = 0;

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }
  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] = sendbuf[startelem[rank]+i];
  }
  curpos = (rank -1 + size) % size;
  for (j=0;j<recvcounts[curpos];j++) {
    tmpbuf[j] = sendbuf[startelem[curpos]+j];
  }

  // Get all the nodes send the data of the node before it.

  for (i=0;i<size-1;i++) {
    MPI_Isend(tmpbuf, recvcounts[curpos], MPI_DOUBLE, 
              (rank+1)%size,
              tag, MPI_COMM_WORLD,&request); 
    //MPI_Wait(&request, &status);

    // curpos indicates the destination of node receiving the data

    curpos = (curpos - 1 + size) % size;

    MPI_Irecv(tmpbuf2, recvcounts[curpos], 
              MPI_DOUBLE, (rank-1+size)%size, 
              tag, MPI_COMM_WORLD, &request); 
    MPI_Wait(&request, &status);

    // for the data received, add its own data and send it on

    for (j=0;j<recvcounts[curpos];j++) {
      tmpbuf[j] = tmpbuf2[j] + sendbuf[startelem[curpos]+j];
    }
  }

  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] += tmpbuf2[i];
  } 
  
  // print results if -o argument defined
  
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 

  return 0;
}

// Fan-in Fan-out Reduce-Scatter (obsolete)

int fan_redscat (double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size, int *recvcounts) {
  int *startelem;
  int *blocksize;
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  double *tmpbuf;
  double *tmpbuf2;
  int sendsize;
  int recvsize;
  int startpt;
  MPI_Status status;
  MPI_Request request;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.  

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  } 

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }  

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }  
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    while(!finished) {
      if (checkrank%2 == 0){
	if ((rank + offset) < size) {
	  MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
		   MPI_COMM_WORLD, &request);
          MPI_Wait(&request, &status);
	  for (i=0;i<args->numelem;i++) {
	    recvbuf[i] += tmpbuf[i];
	  } 
	}
      } else {
	// if my checkrank is odd numbered, I will send my data
	MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
	finished = TRUE;
      }
      
      if (!rank) {
	if (offset*2 >= size) {
	  finished = TRUE;
	}
      }
      
      // Number of nodes to send and receive data cut by half
      checkrank = checkrank/2;
      // offset between ranks is doubled
      offset = offset*2;
    }
  }
  
  // Fan-out

  finished = FALSE;

  // offset keeps track of current offset (initially number of nodes)

  while (!finished) {
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        //rank%offset : send(rank+offset/2);
        sendsize = 0;
        for (i=0;i<offset/2;i++) {
          sendsize += recvcounts[rank+offset/2+i];
        }
        startpt = startelem[rank + offset/2] - startelem[rank];
        for (i=0;i<sendsize;i++) {
          tmpbuf[i] = recvbuf[startpt+i];
        }
        
        MPI_Isend(tmpbuf, sendsize, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
      }
    }
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        //((rank-offset/2)%offset) : recv(rank-offset/2);
        recvsize = 0;
        for (i=0;i<offset/2;i++) {
          recvsize += recvcounts[rank+i];
        }
        MPI_Irecv(tmpbuf2, recvsize, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
      }
      for (i=0;i<recvsize;i++) {
        recvbuf[i] = tmpbuf2[i];
      }
    }
    //if (!rank) printf("gothere\n");
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
    
  }

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}    

// Bi-Directional Exchange Reduce-Scatter (obsolete)

int bidir_redscat (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int * recvcounts) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  double * tmpbuf;
  MPI_Status status;
  MPI_Request request;
  int tag=0;
  int * startelem;

  send = 0;
  offset = 1;
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  startelem = (int*)malloc(size*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  while(!finished) {
    
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }
    
    if(sender) {
      MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &request);      
    } else {
      MPI_Isend(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &request);
    }
    
    MPI_Wait(&request, &status);
    tag++;

    for(i=0;i<args->numelem;i++) {
      recvbuf[i] += tmpbuf[i];
    }   

    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }
 
  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  
  return 0;

}

// Recursive-Halving Reduce-Scatter

int rec_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int *recvcounts) {
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int *blocksize;
  int i, j;
  int sendsize;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curnumelem;
  int curnumrecelem;
  double *tmpbuf;
  int found;
  int tmprank;
  int *realranks;

  MPI_Status status;
  MPI_Request request;  

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  realranks = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));  

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.  

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }  

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }  

  j=0;
  found = FALSE;

  // This portion is a bit dodgy, the position of the final data is calculated
  // by doing a mini-recursive-halving. This seems to be the only way to
  // calculate this value and it does not affect performance. Though a more
  // elegant method should be used.

  while((j<size)&&(!found)) {
    activepos = 0;
    sendsize = size/2;
    send = 0;
    offset = 1;  
    done = FALSE;
    sender = FALSE;

    // Create sender and non-sender pairs.

    while (!done) {
      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((j == i) && send) {
          sender = TRUE;
        }
      }
      
      // If non-sender, swap the active and not active positions. (sender 
      // initally sends the top portion and non-senders send the bottom
      // portion

      if (sender) {
        notactivepos = activepos+sendsize;
      } else {
        notactivepos = activepos;
        activepos = activepos + sendsize;
      }
      
      // There is no send as this is only to calculate the final node send

      if (sendsize == 1) {
        done = TRUE;
      } else {
        activepos = notactivepos;
        sendsize = sendsize / 2;
        offset = offset * 2;
        sender = 0;
        send = 0;    
      }
    }

    // recreate virtual ranks such that the real rank corresponds to the final
    // complete data. i.e. for 8 processors virtual rank 7 holds data for block
    // 0 so has an actual rank of 0.

    if (notactivepos == rank) {
      tmprank = j;
    }
    realranks[notactivepos] = j;
    j++;
  }

  // re-initialise data again
  
  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;
  sender = FALSE;
  done = FALSE;

  // Now the code does the send

  while (!done) {
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((tmprank == i) && send) {
        sender = TRUE;
      }
    }

    // Again pair off sender and non-sender pairs but this time based on the
    // virtual rank

    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    // calculate the size of the block to send

    curnumelem = 0;

    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }

    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    // calculate the size of the block to receive

    curnumrecelem = 0;

    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }

    // send to the node with the tmprank (using their real rank)

    if (sender) {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, realranks[tmprank+offset], 
                tag, 
                MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, realranks[tmprank+offset], 
                tag, 
                MPI_COMM_WORLD, &request);
    } else {
      MPI_Isend(tmpbuf2, curnumelem, MPI_DOUBLE, realranks[tmprank-offset], 
                tag, 
               MPI_COMM_WORLD, &request);
      MPI_Irecv(tmpbuf, curnumrecelem, MPI_DOUBLE, realranks[tmprank-offset], 
                tag, 
                MPI_COMM_WORLD, &request);
    }
    
    MPI_Wait(&request, &status);
    
    // receive, combine and store the information

    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] += tmpbuf[i];
    }

    if (sendsize == 1) {
      done = TRUE;
    } else {
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }

  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }

  // print the result if -o argument defined

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<recvcounts[rank];i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;

}

// Bucket Reduce-Scatter (obsolete)

int buck_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int *recvcounts) {
  
  int *blocksize;
  int *startelem;
  double *tmpbuf2;
  int i, j;
  int tag = 0;
  int cursendpos;
  int currecpos;
  int numiter;
  int count=0;
  double *tmpbuf;
  MPI_Request request;
  MPI_Status status;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }
  
  startelem[0] = 0;
  
  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  } 

  cursendpos = rank - 1;
  
  if (!rank) cursendpos = size - 1;
  
  if ((cursendpos - 1) >= 0) {
    currecpos = cursendpos - 1;
  } else {
    currecpos = size - 1;
  }
  
  numiter = size - 1;

  for(i=0;i<numiter;i++) {
    
    for(j=0;j<blocksize[cursendpos];j++) {
      tmpbuf2[j] = recvbuf[startelem[cursendpos]+j];    
      
    }

    MPI_Isend(tmpbuf2, blocksize[cursendpos], MPI_DOUBLE, (rank+1)%size, 
              tag, MPI_COMM_WORLD,&request);
    
    if (!rank) {
      MPI_Irecv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &request);
    } else {
      MPI_Irecv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &request);
    }
    tag++;

    MPI_Wait(&request, &status);
    
    for(j=0;j<blocksize[currecpos];j++) {
      recvbuf[startelem[currecpos]+j] += tmpbuf[j];
    }

    if ((cursendpos - 1) >= 0) {
      cursendpos--;
    } else {
      cursendpos = size - 1;
    }

    if ((cursendpos - 1) >= 0) {
      currecpos = cursendpos - 1;
    } else {
      currecpos = size - 1;
    }
    count++;
  }

  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<recvcounts[rank];i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf2);
  return 0;

}

// MPI Implementation Reduce-Scatter

int mpi_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int *recvcounts) {
  int i;

  MPI_Reduce_scatter(sendbuf, recvbuf, recvcounts, MPI_DOUBLE, MPI_SUM, 
                MPI_COMM_WORLD);

  if((rank==args->testrank) && (args->output)){
    for(i=0;i<recvcounts[rank];i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}

// Pipeline and Repeated Binary Tree Reduce-Scatter

int pipe_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int * recvcounts) {
  int i, j;
  double * tmpbuf;
  double * tmpbuf2;
  int tag=0;
  MPI_Request request;
  MPI_Status status;
  int * startelem;
  int myrank, finished, offset, checkrank;
  //int count;

  startelem = (int*)malloc(size*sizeof(double));

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  startelem[0] = 0;

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }

  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] = sendbuf[startelem[rank]+i];
  }

  if (args->ring == TRUE) {
    // Pipeline Reduce-Scatter
    
    for (i=0;i<size;i++) {
      if (rank == i) {

	// Send the block for rank-1 node to the next node
        for (j=0;j<recvcounts[(rank-1+size)%size];j++) {
          tmpbuf[j] = sendbuf[startelem[(rank-1+size)%size]+j];
        }
        MPI_Isend(tmpbuf, recvcounts[(rank-1+size)%size], MPI_DOUBLE, 
                  (rank+1)%size,
                  tag, MPI_COMM_WORLD,&request); 
        MPI_Wait(&request, &status);
      } else {
        // Receive the data combine with my own local data and pass it on.
        MPI_Irecv(tmpbuf, recvcounts[(i-1+size)%size], 
                  MPI_DOUBLE, (rank-1+size)%size, 
                  tag, MPI_COMM_WORLD, &request); 
        MPI_Wait(&request, &status);
             
        if (((i - 1 + size) % size) == rank) {
          for (j=0;j<recvcounts[rank];j++) { 
            recvbuf[j] = tmpbuf[j] + sendbuf[startelem[rank]+j]; 
          }
        } else {
          for (j=0;j<recvcounts[(i-1+size)%size];j++) { 
            tmpbuf[j] += sendbuf[startelem[(i-1+size)%size]+j]; 
          }
	  // Once I have sent all the data I have received, I send my own.
          MPI_Isend(tmpbuf, recvcounts[(i-1+size)%size], 
                    MPI_DOUBLE, (rank+1)%size,
                    tag, MPI_COMM_WORLD,&request); 
        }
      }
    }
    
  } else {
    // Repeated Binary Tree
    if (size > 1) {
      // initialise tree traversing variables
      // each node does an inverted binary tree gather
      for (i=0;i<size;i++) {
        myrank = (rank + size - i) % size;
        finished = FALSE;
        offset = 1;
        checkrank = myrank;
        
        // Fan-in
        
       for (j=0;j<recvcounts[i];j++) {
         tmpbuf[j] = sendbuf[startelem[i]+j];
       } 
       
        while(!finished) {
	  // checkrank is used to create sender non-sender pairs. (odd send,
	  // even receive)
          if (checkrank%2 == 0){
	    // receive data
            if ((myrank + offset) < size) {
              MPI_Irecv(tmpbuf2, recvcounts[i], MPI_DOUBLE, 
                        (rank+offset)%size, tag, 
                        MPI_COMM_WORLD, &request);
              MPI_Wait(&request, &status);
              for (j=0;j<recvcounts[i];j++) {
                tmpbuf[j] += tmpbuf2[j];
              } 
            }
          } else {
            // if my checkrank is odd numbered, I will send my data
            MPI_Isend(tmpbuf, recvcounts[i], MPI_DOUBLE, 
                      (rank+size-offset)%size, tag, 
                      MPI_COMM_WORLD, &request);
            MPI_Wait(&request, &status);
            finished = TRUE;
          }
          
          if (!myrank) {
            if (offset*2 >= size) {
              for (j=0;j<recvcounts[i];j++) {
                recvbuf[j] = tmpbuf[j];
              } 
              finished = TRUE;
            }
          }
          
          // Number of nodes to send and receive data cut by half
          checkrank = checkrank/2;
          // offset between ranks is doubled
          offset = offset*2;
        }
      }
    }
  }

  // print results if -o argument defined
  
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 
  
  return 0;
}

// Full Fan-in Reduce-Scatter

int bing_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size, int * recvcounts) {

  int i,j,k;
  double * tmpbuf;
  double * tmpbuf2;
  int tag=0;
  MPI_Request request;
  MPI_Status status;
  int * startelem;

  startelem = (int*)malloc(size*sizeof(double));
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }
  
  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] = sendbuf[startelem[rank]+i];
  }
  
  // i defines the node receiving

  for (i=0;i<size;i++) {
    // the receiving node receives from all the other nodes
    for(j=1;j<size;j++) {
      if (rank == i) {
        MPI_Irecv(tmpbuf2, recvcounts[rank], MPI_DOUBLE, 
                  (rank + j) % size, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
        for (k=0;k<recvcounts[rank];k++) {
          recvbuf[k] += tmpbuf2[k];
        }
      } 
      
      // all other nodes send their data
      if (rank == ((i + j) % size)) {
        for (k=0;k<recvcounts[(rank-j+size)%size];k++) {
          tmpbuf[k] = sendbuf[startelem[(rank-j+size)%size]+k];
        }
	
        // sd(rank, (rank-j+size)%size);
        MPI_Isend(tmpbuf, recvcounts[(rank-j+size)%size], MPI_DOUBLE, 
                  (rank-j+size)%size, tag, 
                  MPI_COMM_WORLD, &request);
        MPI_Wait(&request, &status);
        // send to i-j + size % size
      }
    }
  }

  // print results if -o argument defined

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 
  return 0;

}

// Full Fan-out Reduce-Scatter

int naive_redscat(double *sendbuf, double *recvbuf, struct args *args, int rank, int size, int * recvcounts) {
  int i,j;
  double * tmpbuf;
  double * tmpbuf2;
  int tag=0;
  MPI_Request request;
  MPI_Status status;
  int * startelem;

  startelem = (int*)malloc(size*sizeof(double));
  
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  // The data is divided into blocks and the start of each block is the data
  // for its equilavent node. i.e. startelement[0] is the data block for 
  // node 0.
  
  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }
  
  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] = sendbuf[startelem[rank]+i];
  }
  
  for(i=1;i<size;i++) {
    // send to (rank + i) mod size
    for(j=0;j<recvcounts[(rank + i) % size];j++) {
      tmpbuf[j] = sendbuf[startelem[(rank + i) % size]+j];
    }
    MPI_Isend(tmpbuf, recvcounts[(rank + i)%size], 
	      MPI_DOUBLE, (rank + i)%size, tag, MPI_COMM_WORLD, 
	      &request);
  }

  // all nodes send data to every other node.

  for(i=1;i<size;i++) {
    // receive from (rank + i) mod size
    MPI_Irecv(tmpbuf2, recvcounts[rank], MPI_DOUBLE, 
	      (rank + i)%size,
	      tag, MPI_COMM_WORLD, &request);
    MPI_Wait(&request, &status);
    for (j=0;j<recvcounts[rank];j++) {
      recvbuf[j] += tmpbuf2[j];
    }
  }

  // print results if -o argument defined
  
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 

  return 0;
}

// BLOCKING REDUCE_SCATTER ALGORITHMS (OBSOLETE)

//--------------------------------------------------


int b_ring_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size, int * recvcounts) {
  int tag = 0;
  int i,j;
  double *tmpbuf;
  MPI_Status status;
  int * startelem;

  startelem = (int*)malloc(size*sizeof(double));

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] = sendbuf[j];
    tmpbuf[j] = 0;
  }
  
  
  
  if (!rank) {
    MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, MPI_COMM_WORLD, 
              &status);
    MPI_Send(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
            MPI_COMM_WORLD);
  } else {
    MPI_Send(sendbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
            MPI_COMM_WORLD);
    MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, MPI_COMM_WORLD, 
              &status);
  }

  for(j=0;j<args->numelem;j++) {
    recvbuf[j] += tmpbuf[j];
  }
  
  for(i=2;i<size;i++) {
    

    if (!rank) {
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &status);   
      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
             MPI_COMM_WORLD);
    } else {
      MPI_Send(tmpbuf, args->numelem, MPI_DOUBLE, (rank+1)%size, tag, 
             MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &status);
    } 
    
    tag = i;
    
    for(j=0;j<args->numelem;j++) {
      recvbuf[j] += tmpbuf[j];
    }
  } 
  
   startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  } 
  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 

  return 0;
}

// ! NEED WORK HERE !

int fourway_redscat(double *sendbuf, double *recvbuf, struct args *args, 
		 int rank, int size, int * recvcounts) {
  if (!rank) printf("4 way reduce scatter\n");
  return 0;
}

int b_fan_redscat (double *sendbuf, double *recvbuf, struct args *args, 
                 int rank, int size, int *recvcounts) {
  int *startelem;
  int *blocksize;
  int i; 
  int tag=0;
  int finished;
  int offset;
  int checkrank;
  double *tmpbuf;
  double *tmpbuf2;
  int sendsize;
  int recvsize;
  int startpt;
  MPI_Status status;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  } 

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }  

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }  
  
  if (size > 1) {
    // initialise tree traversing variables
    finished = FALSE;
    offset = 1;
    checkrank = rank;

    // Fan-in

    while(!finished) {
      if (checkrank%2 == 0){
	if ((rank + offset) < size) {
	  MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
		   MPI_COMM_WORLD, &status);
	  for (i=0;i<args->numelem;i++) {
	    recvbuf[i] += tmpbuf[i];
	  } 
	}
      } else {
	// if my checkrank is odd numbered, I will send my data
	MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                  MPI_COMM_WORLD);
	finished = TRUE;
      }
      
      if (!rank) {
	if (offset*2 >= size) {
	  finished = TRUE;
	}
      }
      
      // Number of nodes to send and receive data cut by half
      checkrank = checkrank/2;
      // offset between ranks is doubled
      offset = offset*2;
    }
  }
  
  // Fan-out

  finished = FALSE;

  // offset keeps track of current offset (initially number of nodes)

  while (!finished) {
    if (rank%offset == 0) {
      if ((rank+offset/2) < size) {
        //rank%offset : send(rank+offset/2);
        sendsize = 0;
        for (i=0;i<offset/2;i++) {
          sendsize += recvcounts[rank+offset/2+i];
        }
        startpt = startelem[rank + offset/2] - startelem[rank];
        for (i=0;i<sendsize;i++) {
          tmpbuf[i] = recvbuf[startpt+i];
        }
        
        MPI_Send(tmpbuf, sendsize, MPI_DOUBLE, rank+offset/2, tag, 
                 MPI_COMM_WORLD);
      }
    }
    if((rank-offset/2)%offset == 0) {
      if ((rank-offset/2) >= 0) {
        //((rank-offset/2)%offset) : recv(rank-offset/2);
        recvsize = 0;
        for (i=0;i<offset/2;i++) {
          recvsize += recvcounts[rank+i];
        }
        MPI_Recv(tmpbuf2, recvsize, MPI_DOUBLE, rank-offset/2, tag, 
                 MPI_COMM_WORLD, &status);
      }
      for (i=0;i<recvsize;i++) {
        recvbuf[i] = tmpbuf2[i];
      }
    }
    //if (!rank) printf("gothere\n");
    if (offset == 1) {
      finished = TRUE;
    } else {
      offset = offset/2;
    }
    
  }

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }

  return 0;
}    


int b_bidir_redscat (double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int * recvcounts) {

  int offset;
  int finished = FALSE;
  int send;
  int sender = FALSE;
  int i;
  double * tmpbuf;
  MPI_Status status;
  int tag=0;
  int * startelem;

  send = 0;
  offset = 1;
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  startelem = (int*)malloc(size*sizeof(double));

  for (i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  while(!finished) {
    
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((rank == i) && send) {
        sender = TRUE;
      }
    }
    
    if(sender) {
      MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank+offset, tag, 
                MPI_COMM_WORLD, &status);
      
      
    } else {
      MPI_Recv(tmpbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
                MPI_COMM_WORLD, &status);
      MPI_Send(recvbuf, args->numelem, MPI_DOUBLE, rank-offset, tag, 
               MPI_COMM_WORLD);
    }
    
    tag++;

    for(i=0;i<args->numelem;i++) {
      recvbuf[i] += tmpbuf[i];
    }   

    offset *= 2;
    if(offset >= size) {
      finished = TRUE;
    } else {
      sender = FALSE;
    }
  } 

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  } 
  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }
  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  }
  
  return 0;

}





int b_rec_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int *recvcounts) {
  int *startelem;
  int done = FALSE;
  int offset;
  int send = 0;
  int sender = FALSE;
  int *blocksize;
  int i, j;
  int sendsize;
  double *tmpbuf2;
  int tag = 0;
  int activepos;
  int notactivepos;
  int curnumelem;
  int curnumrecelem;
  double *tmpbuf;
  int found;
  int tmprank;
  int *realranks;
  MPI_Status status;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  realranks = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));  

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }  

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  }  

  j=0;
  found = FALSE;

  while((j<size)&&(!found)) {
    activepos = 0;
    sendsize = size/2;
    send = 0;
    offset = 1;  
    done = FALSE;
    sender = FALSE;

    while (!done) {
      for (i=0;i<size;i++) {
        if (!(i % offset)) {
          send ^= 1;
        }
        if ((j == i) && send) {
          sender = TRUE;
        }
      }
      
      if (sender) {
        notactivepos = activepos+sendsize;
      } else {
        notactivepos = activepos;
        activepos = activepos + sendsize;
      }
      
      if (sendsize == 1) {
        done = TRUE;
      } else {
        activepos = notactivepos;
        sendsize = sendsize / 2;
        offset = offset * 2;
        sender = 0;
        send = 0;    
      }
    }

    if (notactivepos == rank) {
      tmprank = j;
    }
    realranks[notactivepos] = j;
    j++;
  }
  
  activepos = 0;
  sendsize = size/2;
  send = 0;
  offset = 1;
  sender = FALSE;
  done = FALSE;

  while (!done) {
    for (i=0;i<size;i++) {
      if (!(i % offset)) {
        send ^= 1;
      }
      if ((tmprank == i) && send) {
        sender = TRUE;
      }
    }

    if (sender) {
      notactivepos = activepos+sendsize;
    } else {
      notactivepos = activepos;
      activepos = activepos + sendsize;
    }

    curnumelem = 0;

    for (i=activepos; i< activepos + sendsize;i++) {
      curnumelem = curnumelem + blocksize[i];
    }

    for (i=0;i<curnumelem;i++) {
      tmpbuf2[i] = recvbuf[i+startelem[activepos]];
    } 

    curnumrecelem = 0;

    for (i=notactivepos; i< notactivepos + sendsize; i++) {
      curnumrecelem = curnumrecelem + blocksize[i];
    }

    if (sender) {
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, realranks[tmprank+offset], 
                tag, 
                MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, realranks[tmprank+offset], 
                tag, 
                MPI_COMM_WORLD, &status);
    } else {
      MPI_Send(tmpbuf2, curnumelem, MPI_DOUBLE, realranks[tmprank-offset], 
                tag, 
               MPI_COMM_WORLD);
      MPI_Recv(tmpbuf, curnumrecelem, MPI_DOUBLE, realranks[tmprank-offset], 
                tag, 
                MPI_COMM_WORLD, &status);
    }
 
    
    for(i=0;i<curnumrecelem;i++) {
      recvbuf[i+startelem[notactivepos]] += tmpbuf[i];
    }

    if (sendsize == 1) {
      done = TRUE;
    } else {
      activepos = notactivepos;
      sendsize = sendsize / 2;
      offset = offset * 2;
      sender = 0;
      send = 0;    
    }
  }

   for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }
  if((rank==args->testrank) && (args->output)){
    for(i=0;i<recvcounts[rank];i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  return 0;

}

int b_buck_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int *recvcounts) {
  
  int *blocksize;
  int *startelem;
  double *tmpbuf2;
  int i, j;
  int tag = 0;
  int cursendpos;
  int currecpos;
  int numiter;
  int count=0;
  double *tmpbuf;
  MPI_Status status;

  blocksize = (int*)malloc(size*sizeof(int));
  startelem = (int*)malloc(size*sizeof(int));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf = (double*)malloc(args->numelem*sizeof(double));

  for(i=0;i<args->numelem;i++) {
    recvbuf[i] = sendbuf[i];
  }

  for (i=0;i<size;i++) {
    blocksize[i] = recvcounts[i];
  }
  
  startelem[0] = 0;
  
  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + blocksize[i-1];
  } 

  cursendpos = rank - 1;
  
  if (!rank) cursendpos = size - 1;
  
  if ((cursendpos - 1) >= 0) {
    currecpos = cursendpos - 1;
  } else {
    currecpos = size - 1;
  }
  
  numiter = size - 1;

  for(i=0;i<numiter;i++) {
    
    for(j=0;j<blocksize[cursendpos];j++) {
      tmpbuf2[j] = recvbuf[startelem[cursendpos]+j];    
      
    }

    MPI_Send(tmpbuf2, blocksize[cursendpos], MPI_DOUBLE, (rank+1)%size, 
              tag, MPI_COMM_WORLD);
    
    if (!rank) {
      MPI_Recv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, size-1, tag, 
                MPI_COMM_WORLD, &status);
    } else {
      MPI_Recv(tmpbuf, blocksize[currecpos], MPI_DOUBLE, rank-1, tag, 
                MPI_COMM_WORLD, &status);
    }
    tag++;
    
    for(j=0;j<blocksize[currecpos];j++) {
      recvbuf[startelem[currecpos]+j] += tmpbuf[j];
    }

    if ((cursendpos - 1) >= 0) {
      cursendpos--;
    } else {
      cursendpos = size - 1;
    }

    if ((cursendpos - 1) >= 0) {
      currecpos = cursendpos - 1;
    } else {
      currecpos = size - 1;
    }
    count++;
  }
  for (i=0;i<recvcounts[rank];i++) {
    tmpbuf[i] = recvbuf[startelem[rank]+i];
    recvbuf[i] = tmpbuf[i];
  }
  if((rank==args->testrank) && (args->output)){
    for(i=0;i<recvcounts[rank];i++) {
      printf(" %f",recvbuf[i]);
    }
    printf("\n");
  }

  free(startelem);
  free(blocksize);
  free(tmpbuf2);
  return 0;

}

int b_pipe_redscat(double *sendbuf, double *recvbuf, struct args *args, 
                  int rank, int size, int * recvcounts) {
  int i, j, k;
  double * tmpbuf;
  double * tmpbuf2;
  int tag=0;
  MPI_Status status;
  int * startelem;

  startelem = (int*)malloc(size*sizeof(double));

  tmpbuf = (double*)malloc(args->numelem*sizeof(double));
  tmpbuf2 = (double*)malloc(args->numelem*sizeof(double));

  startelem[0] = 0;

  for (i=1;i<size;i++) {
    startelem[i] = startelem[i-1] + recvcounts[i-1];
  }

  for (i=0;i<recvcounts[rank];i++) {
    recvbuf[i] = sendbuf[startelem[rank]+i];
  }
  for (i=0;i<size;i++) {
    for (j=0;j<size;j++) {
      if (j != i) {
        if (rank == i) {
          // re(rank);
          MPI_Recv(tmpbuf, recvcounts[i], MPI_DOUBLE, j, tag, 
                    MPI_COMM_WORLD, &status); 
          for(k=0;k<recvcounts[i];k++) {
            recvbuf[k] += tmpbuf[k];
          }
          
        }
        if (rank == j) {
          // sd(rank);
          for(k=0;k<recvcounts[i];k++) {
            tmpbuf2[k] = sendbuf[startelem[i]+k];
          }
          MPI_Send(tmpbuf2, recvcounts[i], MPI_DOUBLE, i, 
                    tag, MPI_COMM_WORLD); 
        }
      }
      
    }
  }

  if((rank==args->testrank) && (args->output)){
    for (i=0;i<recvcounts[rank];i++) {
      printf(" %f ", recvbuf[i]);
    }
    printf("\n");
  } 

  return 0;
}







